import re
from dentmark import defs_manager, TagDef, OptionalUnique, RequiredUnique

CONFIG_TAG_SET = 'indentgen_config'
SUBSITE_CONFIG_TAG_SET = 'indentgen_subsite_config'

config_tag_set = defs_manager.get_tag_set(CONFIG_TAG_SET)


class ConfigURLContext(TagDef):
    is_context = True

    min_num_text_nodes = 1
    max_num_text_nodes = 1

    url_pattern = re.compile(r'^[a-z0-9](?:[a-z0-9-]*[a-z0-9])?$')

    def validate(self):
        val = self.get_data()
        if not self.url_pattern.match(val):
            return f"Tag '{self.tag_name}' contains invalid format for a slug. Can contain only [a-z0-9-] and cannot begin or end with '-'"



@config_tag_set.register()
class ConfigRoot(TagDef):
    tag_name = 'root'

    min_num_text_nodes = 0
    max_num_text_nodes = 0

    def render_main(self):
        return self.context # just return the context to get the config data so we can use render. Config file has no body content, all context

@config_tag_set.register()
class ConfigDateArchiveURL(ConfigURLContext):
    tag_name = 'date_archive_url'

    parents = [OptionalUnique('root')]


@config_tag_set.register()
class ConfigSingleParam(TagDef):
    tag_name = 'title'
    is_context = True

    min_num_text_nodes = 1
    max_num_text_nodes = 1

    parents = [OptionalUnique('root')]


@config_tag_set.register()
class ConfigDescription(ConfigSingleParam):
    tag_name = 'description'


@config_tag_set.register()
class ConfigPerPage(ConfigSingleParam):
    tag_name = 'per_page'


    def process_data(self, data):
        per_page = int(data[0])
        if per_page < 0:
            raise ValueError
        return per_page


    def validate(self):
        try:
            self.get_data()
        except ValueError:
            return f"Tag '{self.tag_name}' in config expects a positive integer"


# copy from CONFIG_TAG_SET and add a few more required tags
subsite_config_tag_set = defs_manager.copy_tag_set(SUBSITE_CONFIG_TAG_SET, CONFIG_TAG_SET)

@subsite_config_tag_set.register()
class SubsiteConfigParentSlug(TagDef):
    tag_name = 'parent_slug'
    is_context = True

    min_num_text_nodes = 1
    max_num_text_nodes = 1

    parents = [RequiredUnique('root')]


@subsite_config_tag_set.register()
class SubsiteConfigSubsiteSlug(TagDef):
    tag_name = 'subsite_slug'
    is_context = True

    min_num_text_nodes = 1
    max_num_text_nodes = 1

    parents = [RequiredUnique('root')]

@subsite_config_tag_set.register()
class SubsiteConfigTemplatePathPrefix(TagDef):
    tag_name = 'template_path_prefix'
    is_context = True

    min_num_text_nodes = 1
    max_num_text_nodes = 1

    parents = [OptionalUnique('root')]



from indentgen.page_store import PageStore
from calendar import month_name
from pathlib import Path

PAGE_URL = 'page' #TODO make this configurable in site settings?

class Endpoint:
    use_template = 'pages/home.html'
    srp = None
    is_taxonomy = False
    add_to_sitemap = True
    is_gallery = False

    def __init__(self, indentgen_obj, url_components, page=None, subsite_config=None):
        self.url_components = url_components
        self.page = page
        self.indentgen = indentgen_obj
        self.subsite_config = subsite_config
        self.child_pages = None
        self.paginator_page = None


    def get_output_path(self):
        if self.page is None or self.page == 0:
            return Path(*self.url_components)
        else:
            components = list(self.url_components) # components could be tuples
            components.append(PAGE_URL)
            components.append(str(self.page + 1))
            return Path(*components)

    @property
    def url(self):
        if not self.url_components and not self.page:
            return '/'
        url = '/'.join(self.url_components)
        if self.page is not None and self.page > 0:
            url += f'/{PAGE_URL}/{self.page + 1}'
        # all routes except home are xxx/xxx and need to
        # have the '/' added at he beginning and end.
        # Paginated homepage URLS are just /page/xx
        # and so will have //page/xx if this check is
        # not in place. Also prepends / to /404.html,
        # /sitemap.xml etc.
        prepend_slash = '/' if url[0] != '/' else ''
        append_slash = '/' if not (len(self.url_components) == 1 and '.' in url) else '' # don't append if 404.html, sitemap.xml etc.
        return f"{prepend_slash}{url}{append_slash}"

    @property
    def identifier(self):
        # TODO probably shouldn't reference self.srp here since it's defined by
        # a subsclass
        return self.srp if self.srp is not None else self.url

    @property
    def meta(self):
        return {} # for compatibility with templates rather that having to do hasattr checks


    def render(self):
        print('rendering', self.url)
        context = {
            'page': self,
            'config': self.subsite_config or self.indentgen.config
        }

        use_template = self.use_template
        if self.subsite_config:
            template_path_prefix = self.subsite_config.get('template_path_prefix', '').strip('/') # strip any leading or trailing '/'
            if template_path_prefix:
                use_template = f'{template_path_prefix}/{self.use_template}'

        template = self.indentgen.templates.get_template(use_template)
        return template.render(**context)


    def next_page(self):
        next_ep = Endpoint(self.indentgen, self.url_components, self.page + 1, self.subsite_config)
        next_ep.child_pages = self.child_pages
        return next_ep


class ContentEndpoint(Endpoint):
    use_template = 'pages/content.html'

    def __init__(self, indentgen_obj, url_components, page, srp, subsite_config=None):
        super().__init__(indentgen_obj, url_components, page, subsite_config)
        self.srp = srp
        self.prev = None
        self.next = None

        # change the template if use_template is set in meta
        use_template = self.meta.get('use_template')
        if use_template:
            self.use_template = use_template

    #TODO perhaps optimize this so it doesn't read from cached wisdom .html files if only fetching root
    def get_rendered(self):
        return self.indentgen.wisdom.get_rendered(self.srp, self.is_taxonomy)

    @property
    def root(self):
        return self.get_rendered()[1]

    @property
    def content(self):
        return self.get_rendered()[0]

    @property
    def context(self):
        return self.root.context

    @property
    def meta(self):
        return self.context['meta']

    @property
    def summary(self):
        meta_summary = self.root.context['meta'].get('summary_content', '')
        if meta_summary:
            return meta_summary
        else:
            inline_summary_list = self.root.collectors.get('sum', [])
            if inline_summary_list:
                return ''.join(inline_summary_list)
        return ''

    @property
    def taxonomies(self):
        return self.meta.get('taxonomy', {})

    @property
    def title(self):
        return self.meta.get('title', '')

    @property
    def description(self):
        return self.meta.get('description', '')

    @property
    def slug(self):
        return self.meta['slug']

    @property
    def is_gallery(self):
        return self.paginator_page is not None

    def get_taxonomy_group(self, top_level_taxonomy):
        #TODO maybe cache this in indentgen so we're not having to re-filter these over and over for every hit
        filter_page_taxonomies = [k for k,v in self.indentgen.taxonomy_map.items() if v['top_level'] == top_level_taxonomy]
        taxonomies_for_page =  set(filter_page_taxonomies).intersection(self.taxonomies)
        return PageStore([self.indentgen.taxonomy_map[slug]['endpoint'] for slug in taxonomies_for_page])

    def next_page(self):
        next_ep = ContentEndpoint(self.indentgen, self.url_components, self.page + 1, self.srp, self.subsite_config)
        next_ep.child_pages = self.child_pages
        return next_ep


class ContentGalleryEndpoint(Endpoint):
    use_template = 'pages/gallery_item.html'
    add_to_sitemap = False

    def __init__(self, indentgen_obj, url_components, srp, photo_data, gallery_endpoint): #TODO photogalleries in subsites?
        super().__init__(indentgen_obj, url_components, None, None)
        self.srp = srp # srp of image
        self.photo_data = photo_data
        self.gallery_endpoint = gallery_endpoint
        self.prev = None
        self.next = None

    def next_page(self):
        return None # this class does not have paginated pages, so this method should never be called


class TaxonomyEndpoint(ContentEndpoint):
    use_template = 'pages/taxonomy.html'
    is_taxonomy = True
    is_gallery = False

    @property
    def breadcrumbs(self):
        path_titles = []
        focused = self.indentgen.taxonomy_map[self.slug]
        while 'parent' in focused:
            focused = self.indentgen.taxonomy_map[focused['parent']]
            path_titles.append({'title': focused['title'], 'url': focused['endpoint'].url})
        path_titles.reverse()
        return path_titles

    @property
    def slug(self):
        return self.meta['slug_path']

    def next_page(self):
        next_ep = TaxonomyEndpoint(self.indentgen, self.url_components, self.page + 1, self.srp, self.subsite_config)
        next_ep.child_pages = self.child_pages
        return next_ep


class RedirectEndpoint(Endpoint):
    use_template = 'pages/redirect.html'
    add_to_sitemap = False

    def __init__(self, indentgen_obj, from_url_components, to_endpoint):
        super().__init__(indentgen_obj, from_url_components, None)
        self.to_endpoint = to_endpoint

    @property
    def identifier(self):
        return self.to_endpoint.identifier

    def next_page(self):
        return None # this class does not have paginated pages, so this method should never be called


class StaticServeEndpoint(Endpoint):
    use_template = None
    add_to_sitemap = False

    def __init__(self, indentgen_obj, url_components):
        super().__init__(indentgen_obj, url_components, None)

    def render(self):
        return None

    def next_page(self):
        return None # this class does not have paginated pages, so this method should never be called


class CachedImgEndpoint(StaticServeEndpoint):
    add_to_sitemap = False


class DateArchiveEndpoint(Endpoint):
    use_template = 'pages/date_archive.html'
    add_to_sitemap = False

    @property
    def title(self):
        if len(self.url_components) == 1:
            return 'Date Archive'
        year = self.url_components[1]
        month = self.url_components[2] if len(self.url_components) == 3 else ''
        if month:
            month = month_name[int(month)]
        return f'{month} {year}'.strip()

    @property
    def meta(self):
        return {}

    @property
    def summary(self):
        return ''

    @property
    def breadcrumbs(self):
        if len(self.url_components) == 1:
            return []
        archive_url = f"/{self.indentgen.config['date_archive_url']}/"
        path_titles = [{'title': 'Date Archive', 'url': archive_url }]
        if len(self.url_components) == 3:
            path_titles.append({'title': self.url_components[1], 'url': f'{archive_url}{self.url_components[1]}/'})
        return path_titles


# 404, sitemap, rss

class Http404Endpoint(Endpoint):
    use_template = '404.html'
    add_to_sitemap = False

    def __init__(self, indentgen_obj):
        super().__init__(indentgen_obj, ['404.html'])


class RssEndpoint(Endpoint):
    use_template = 'index.xml'
    add_to_sitemap = False

    def __init__(self, indentgen_obj):
        super().__init__(indentgen_obj, ['index.xml'])


class SiteMapEndpoint(Endpoint):
    use_template = 'sitemap.xml'
    add_to_sitemap = False

    def __init__(self, indentgen_obj):
        super().__init__(indentgen_obj, ['sitemap.xml'])




#TODO enforce that series must have part #s

import sys
import shutil
import importlib
from pathlib import Path

import dentmark

from mako.template import Template
from mako.lookup import TemplateLookup

from indentgen.wisdom import Wisdom
from indentgen.path_dict import PathDict
from indentgen.default_definitions.content_tag_defs import CONTENT_TAG_SET
from indentgen.default_definitions.taxonomy_tag_defs import TAXONOMY_TAG_SET
from indentgen.taxonomy_def_set import TaxonomyDefSet
from indentgen.endpoints import PAGE_URL, Endpoint, ContentEndpoint, ContentGalleryEndpoint, TaxonomyEndpoint, RedirectEndpoint, StaticServeEndpoint, CachedImgEndpoint, DateArchiveEndpoint, Http404Endpoint, RssEndpoint, SiteMapEndpoint
from indentgen.paginator import Paginator
from indentgen.page_store import PageStore


class Indentgen:
    WISDOM_DIR = '_wisdom'
    CONTENT_DIR = 'content'
    TAXONOMY_DIR = 'taxonomy'
    CUSTOM_DEFS_MODULE_NAME = 'indentgen_defs'
    CONFIG_FILE_NAME = 'config.dentmark'
    THEME_DIR = 'theme'
    TEMPLATE_DIR = Path(THEME_DIR) / 'templates'
    TEMPLATE_CACHE_DIR = Path(WISDOM_DIR) / 'mako_modules'
    STATIC_DIR = Path(THEME_DIR) / 'static'

    #TODO puth these in config.dentmark instead?
    OUTPUT_DIR = 'publish'
    STATIC_URL = '_static'
    CACHE_BUST_STATIC_EXTENSIONS = ('.css', '.js') #TODO put this in config.dentmark instead?
    GALLERY_IMAGE_EXTENSIONS = ('.jpg', '.jpeg', '.png')

    IMAGE_URL = '_img'

    DEFAULT_PER_PAGE = 25
    DEFAULT_PER_PAGE_GALLERY = 50

    def __init__(self, site_path):
        self.site_path = Path(site_path)

        sys.path.append(str(self.site_path))

        try:
            self.defs_module = importlib.import_module(self.CUSTOM_DEFS_MODULE_NAME)
        except ModuleNotFoundError:
            print('No custom definitions found') #TODO better error or warning here?

        self.content_path = self.site_path / self.CONTENT_DIR
        self.taxonomy_path = self.site_path / self.TAXONOMY_DIR
        self.wisdom_path = self.site_path / self.WISDOM_DIR

        self.static_path = self.site_path / self.STATIC_DIR
        self.output_path = self.site_path / self.OUTPUT_DIR
        self.static_output_path = self.output_path / self.STATIC_URL

        self.img_output_path = self.output_path / self.IMAGE_URL

        self.config_file_path = self.site_path / self.CONFIG_FILE_NAME

        template_dir = self.site_path / self.TEMPLATE_DIR
        template_cache_dir = self.site_path / self.TEMPLATE_CACHE_DIR
        self.templates = TemplateLookup(directories=[template_dir], module_directory=template_cache_dir)

        self.wisdom = Wisdom(self)
        self.config = self.wisdom.get_config()

        self.routes = {}

        self.pk_link_map = {}
        self.slug_map = {}
        self.subsite_data = PathDict()

        self._patch_def_sets()

        self._build_taxonomy_map()
        self._check_taxonomy_tags_meta(is_taxonomy=True)

        self._find_subsite_srps()

        # do a pass of the content parsing only the meta to build the PK map so that
        # all pks are known prior to rendering the full page content
        self._pre_populate_meta_pk()

        self._validate_subsite_slugs()

        self._build_page_store()

        self._check_taxonomy_tags_meta(is_taxonomy=False)

        self._generate_taxonomy_pagination_routes()
        self._generate_date_archive_routes()
        self._generate_home_pagination_routes()

        self._generate_404_route()
        self._generate_rss_route()
        self._generate_sitemap_route()

        self._build_static_file_remapping()
        self._build_resized_img_endpoints()


    def _gen_walk_content(self, is_taxonomy=False, meta_only=False):
        use_path = self.taxonomy_path if is_taxonomy else self.content_path
        for f in use_path.glob('**/*.dentmark'):
            srp = f.relative_to(self.site_path)

            # skip over subsite config
            if srp.name == self.CONFIG_FILE_NAME:
                continue

            rendered, root = self.wisdom.get_rendered(srp, is_taxonomy, meta_only)
            yield srp, rendered, root


    def _add_route(self, endpoint):
        collision_endpoint = self.routes.get(endpoint.url)
        if collision_endpoint is not None:
            raise Exception(f"URL conflict for '{endpoint.url}': {endpoint.identifier} and {collision_endpoint.identifier}")
        self.routes[endpoint.url] = endpoint


    def _patch_def_sets(self):
        content_tag_set = dentmark.defs_manager.get_tag_set(CONTENT_TAG_SET)
        taxonomy_tag_set = dentmark.defs_manager.get_tag_set(TAXONOMY_TAG_SET)

        # Monkey patch this to use the extended class to defer resolving tag names
        # Kinda hacky, but it works

        for tag_set in (content_tag_set, taxonomy_tag_set):
            new_def_set = TaxonomyDefSet.copy_from_def_set(tag_set)
            dentmark.defs_manager.def_sets[tag_set.tag_set_name] = new_def_set


    # {tax_slug: {}, tax_slug: {meta.., children: []}}
    def _build_taxonomy_map(self):
        tax_map = {}
        top_level_taxonomies = []

        for srp, rendered, root in self._gen_walk_content(is_taxonomy=True):
            meta = root.context['meta']

            slug_path = meta['slug_path']
            collision = tax_map.get(slug_path)
            if collision:
                raise Exception(f"Taxonomy slugs conflict: {srp} and {collision['srp']}")

            tax_map[slug_path] = {'slug_path': slug_path, 'srp': srp, 'title': meta['title']}
            tax_map[slug_path]['pseudo'] = meta['pseudo'] if 'pseudo' in meta else False
            tax_map[slug_path]['gallery'] = meta['gallery'] if 'gallery' in meta else False

            slug_path_components = slug_path.split('/')

            parent = '/'.join(slug_path_components[:-1])

            #TODO enforce that pseudo taxonomies cannot have children?

            if parent:
                tax_map[slug_path]['parent'] = parent
            else:
                top_level_taxonomies.append(slug_path)


        for slug_path, info in tax_map.items():
            parent_slug_path = info.get('parent')
            if parent_slug_path:
                try:
                    children = tax_map[parent_slug_path].setdefault('children', [])
                except KeyError:
                    raise Exception(f"{info['srp']}: invalid parent taxonomy: '{parent_slug_path}'")
                children.append(slug_path)

            url_components = info['slug_path'].split('/')

            info['top_level'] = url_components[0]

            taxonomies = info.get('taxonomy', [])
            srp = info['srp']
            del info['srp']

            endpoint_0 = TaxonomyEndpoint(self, url_components, 0, srp)

            # add this to taxonomy map
            info['endpoint'] = endpoint_0


        self.taxonomy_map = tax_map
        self.top_level_taxonomies = top_level_taxonomies


    def _check_taxonomy_tags_meta(self, is_taxonomy):
        for srp, rendered, root in self._gen_walk_content(is_taxonomy):
            meta = root.context['meta']
            taxonomy = meta.get('taxonomy')
            if taxonomy is not None:
                invalid = set(taxonomy).difference(self.taxonomy_map)
                if invalid:
                    raise Exception(f"Invalid taxonomy tag(s) {invalid} in meta: {srp}")


    # locate other config files in content dir so that we can handle subsites correctly
    def _find_subsite_srps(self):
        use_path = self.content_path
        for f in use_path.glob(f'**/{self.CONFIG_FILE_NAME}'):
            srp = f.relative_to(self.site_path)
            subsite_dir = srp.parent

            subsite_config = self.wisdom.get_subsite_config(subsite_dir)

            self.subsite_data.add(subsite_dir, {'config': subsite_config, 'config_srp': srp})


    # first pass is to resolve all of the pks in the meta, so that they are available
    # when rendering the full body. This is needed to reslove link url's that are PKs in the dentmark
    def _pre_populate_meta_pk(self):
        for srp, rendered, root in self._gen_walk_content(is_taxonomy=False, meta_only=True):

            slug = root.context['meta']['slug']

            conflicting_slug_srp = None

            # verify that slug does not conflict with a taxonomy slug
            if slug in self.taxonomy_map:
                conflicting_slug_srp = self.taxonomy_map[slug]['endpoint'].srp

            # verify that slug does not conflict with another page or subsite slug
            elif slug in self.slug_map:
                conflicting_slug_srp = self.slug_map[slug]['srp']

            if conflicting_slug_srp:
                raise Exception(f"Slugs conflict. Both are '{slug}'. Must be unique: {srp} and {conflicting_slug_srp}")

            pk = root.context['meta'].get('pk')

            #TODO make this configurable in setting whether to use pk shortcuts or not
            url = f'{pk}-{slug}' if pk is not None else slug

            self.slug_map[slug] = {'slug': slug, 'srp': srp, 'last_url_part': url, 'subsite_data': self.subsite_data.get(srp)}

            if pk:
                self.pk_link_map[pk] = slug


    def _validate_subsite_slugs(self):
        # now that we have ALL page slugs stored in self.slug_map, validate that the subsite slugs and parents make sense
        for subsite_data in self.subsite_data:
            config = subsite_data['config']
            config_srp = subsite_data['config_srp']
            slug = config['subsite_slug']
            parent_slug = config['parent_slug']

            if parent_slug not in self.slug_map:
                raise Exception(f"parent_slug '{parent_slug}' invalid in subsite config: {config_srp}")

            if slug in self.slug_map:
                conflicting_srp = self.slug_map[slug]['srp']
                raise Exception(f"subsite_slug '{slug}' conflicts with existing slug: {config_srp} and {conflicting_srp}")

            self.slug_map[slug] = {'slug': slug, 'srp': config_srp, 'last_url_part': slug, 'subsite_data': subsite_data}


    def _resolve_url_components(self, slug):
        slug_data = self.slug_map[slug]
        components = [slug_data['last_url_part']]

        seek = False
        while slug_data.get('subsite_data'):
            subsite_data = slug_data.get('subsite_data')
            config = subsite_data['config']
            subsite_slug = config['subsite_slug']
            if subsite_slug != slug:
                components.append(subsite_slug)
            seek = True
            slug_data = self.slug_map[config['parent_slug']]

        if seek:
            components.append(slug_data['last_url_part'])

        components.reverse()

        return components


    def _get_content_gallery_endpoint(self, gallery_ctx, url_components, srp, subsite_config):
        endpoint_0 = ContentEndpoint(self, url_components, 0, srp, subsite_config)

        # TODO gather photos. For now, just assume all photos live in same directory (alongside)
        # dentmark file. Maybe later, we can make this search location customizable with a context
        # tag or something.

        child_pages = PageStore()
        image_srps = []
        search_dir = srp.parent
        for f in search_dir.glob(f'**/*.*'):
            for ext in self.GALLERY_IMAGE_EXTENSIONS:
                if ext.lower() == f.suffix.lower():
                    image_srps.append(f)
                    break
        image_srps.sort()

        per_page_gallery = gallery_ctx.get('per_page', self.DEFAULT_PER_PAGE_GALLERY)

        prev = None
        for i,image_srp in enumerate(image_srps):
            # gallery url should be /XX-entry-slug/[image_1_indexed]
            gal_url_components = url_components + [str(i+1)]

            photo_data = {}
            #TODO iterating over this every time isn't efficient. Use sorted dict or something?
            for img_data in gallery_ctx.get('img', []):
                if img_data['key'] == image_srp:
                    photo_data = img_data
                    break

            gal_endpoint = ContentGalleryEndpoint(self, gal_url_components, image_srp, photo_data, endpoint_0) #TODO add attrs later
            if prev is not None:
                prev.next = gal_endpoint
                gal_endpoint.prev = prev
            child_pages.add(gal_endpoint)
            self._add_route(gal_endpoint)
            prev = gal_endpoint

        endpoint_0.child_pages = child_pages
        self._make_paginated_routes(endpoint_0, per_page_gallery)
        return endpoint_0


    def _build_page_store(self):
        page_store = PageStore()
        non_pk_page_store = PageStore()
        for srp, rendered, root in self._gen_walk_content(is_taxonomy=False):
            meta = root.context['meta']

            slug = meta['slug']
            url_components = self._resolve_url_components(slug)

            subsite_config = None
            subsite_data = self.subsite_data.get(srp)
            if subsite_data:
                subsite_config = subsite_data['config']

            has_gallery = False
            for tax_slug_path in meta.get('taxonomy', {}):
                if self.taxonomy_map[tax_slug_path]['gallery']:
                    has_gallery = True
                    break

            if has_gallery:
                gallery_ctx = meta.get('gallery', {})

                endpoint = self._get_content_gallery_endpoint(gallery_ctx, url_components, srp, subsite_config)
            else:
                #None is for 0th page, these won't have multiple pages
                endpoint = ContentEndpoint(self, url_components, None, srp, subsite_config)
                self._add_route(endpoint)

            pk = meta.get('pk')

            use_page_store = page_store
            use_non_pk_page_store = non_pk_page_store

            if subsite_data:
                use_page_store = subsite_data.setdefault('page_store', PageStore())
                use_non_pk_page_store = subsite_data.setdefault('non_pk_page_store', PageStore())

            if pk is not None:
                redirect_endpoint = RedirectEndpoint(self, (str(pk),), endpoint)
                self._add_route(redirect_endpoint)
                use_page_store.add(endpoint)
            else:
                use_non_pk_page_store.add(endpoint)

        page_store.annotate_nav() # adds prev, next attrs to Endpoint objs for page nav

        # do the same for all subsite page stores
        for subsite_data in self.subsite_data:

            # TODO is this a good way to handle this? Should I always combine these for both base site and subsites,
            # and maybe just exclude pages from home listing explicitly by using a pseudo taxonomy or something rather
            # relying on presence/absence of PK?
            subsite_page_store = subsite_data.get('page_store', PageStore())

            # message archive doesn't have any PK pages, so just combine these for now. Not sure if this is good default behavior for every subsite though
            subsite_page_store += subsite_data.get('non_pk_page_store', PageStore())

            subsite_page_store.annotate_nav()

        self.page_store = page_store
        self.non_pk_page_store = non_pk_page_store


    def _make_paginated_routes(self, base_endpoint, per_page=None):
        url_components = base_endpoint.url_components

        base_redirect = RedirectEndpoint(self, url_components + [PAGE_URL] + ['1'], base_endpoint) # 2021/3/page/1 -> 2021/3/
        self._add_route(base_redirect)

        page_redirect = RedirectEndpoint(self, url_components + [PAGE_URL], base_endpoint) # 2021/3/page -> 2021/3/
        self._add_route(page_redirect)

        #TODO this paginator stuff is a mess. clean it up. The PageStore should probably generate the paginator,
        # not all of this attaching the paginator to the endpoint crap. Plus violating DRY doing this in about 3 places

        if per_page is None:
            per_page = self.config.get('per_page', self.DEFAULT_PER_PAGE)

        pag = Paginator(base_endpoint.child_pages, per_page)

        for endpoint in pag.gen_all_pages(base_endpoint):
            self._add_route(endpoint)


    def _has_deep_children(self, tax_slug):
        children_stack = [tax_slug]
        while children_stack:
            focused = self.taxonomy_map[children_stack.pop()]
            has_children = len(focused['endpoint'].child_pages.exclude_taxonomies()) > 0
            if has_children:
                return True

            children_stack.extend(focused.get('children', []))
        return False


    def _generate_taxonomy_pagination_routes(self):
        per_page = self.config.get('per_page', self.DEFAULT_PER_PAGE)

        all_taxonomy_endpoints = PageStore([info['endpoint'] for info in self.taxonomy_map.values()])
        all_endpoints = self.page_store + self.non_pk_page_store + all_taxonomy_endpoints

        # add the subsite pages to all_endpoints
        for subsite_data in self.subsite_data:
            subsite_page_store = subsite_data.get('page_store')
            if subsite_page_store:
                all_endpoints += subsite_page_store

            non_pk_page_store = subsite_data.get('non_pk_page_store')
            if non_pk_page_store:
                all_endpoints += non_pk_page_store

        # first pass to gather all of the child pages for each taxonomy
        for slug, info in self.taxonomy_map.items():
            info['endpoint'].child_pages = all_endpoints.filter_by_topic(slug)

        for slug, info in self.taxonomy_map.items():
            endpoint_0 = info['endpoint']

            # remove the taxonomy children that do not have at least 1 page in their children chain
            #eligible_child_tax_endpoints = []
            for child_slug in info.get('children', []):
                if not self._has_deep_children(child_slug):
                    child_endpoint = self.taxonomy_map[child_slug]['endpoint']
                    endpoint_0.child_pages.remove(child_endpoint)

            child_pages = endpoint_0.child_pages.list_view_sort(slug)
            endpoint_0.child_pages = child_pages

            # skip if the child_pages store is still empty after all that (i.e. no pages, and no eligible child taxonomies with a page in their children)
            # also skip if pseuod
            if info['pseudo'] or not child_pages:
                continue

            self._make_paginated_routes(endpoint_0)


    def _generate_date_archive_routes(self):
        date_archive_url = self.config.get('date_archive_url')
        if not date_archive_url:
            # don't make date archives
            return

        dated_endpoints = self.page_store.only_dated() + self.non_pk_page_store.only_dated()
        by_months = dated_endpoints.group_by_date()

        year_list_endpoint = DateArchiveEndpoint(self, [date_archive_url], 0)
        year_list_endpoint.child_pages = PageStore()

        last_year = None

        for (year, month), page_store in sorted(by_months.items()): # this will iterate through the months in ascending order
            url_components = [date_archive_url, str(year), str(month)]

            month_0 = DateArchiveEndpoint(self, url_components, 0)

            # archive pages should be in ascending order
            month_0.child_pages = page_store.order_by_date(descending=False)

            self._make_paginated_routes(month_0)

            if last_year != year or not last_year:
                year_archive_endpoint = DateArchiveEndpoint(self, [date_archive_url, str(year)], 0)
                year_archive_endpoint.child_pages = PageStore()
                year_list_endpoint.child_pages.add(year_archive_endpoint)

            year_list_endpoint.child_pages[-1].child_pages.add(month_0)

            last_year = year

        # make year pages
        for year_endpoint in year_list_endpoint.child_pages:
            self._make_paginated_routes(year_endpoint)

        # make date archive root
        self._make_paginated_routes(year_list_endpoint)


    def _generate_home_pagination_routes_helper(self, per_page, page_store, prefix_components, subsite_config=None):
        endpoint_home_0 = Endpoint(self, prefix_components, 0, subsite_config)

        child_pages = page_store.recent()
        pag = Paginator(child_pages, per_page)

        endpoint_home_0.child_pages = child_pages # attach this in addition to paginator, for completeness although home probably won't need it

        base_redirect = RedirectEndpoint(self, prefix_components + [PAGE_URL] + ['1'], endpoint_home_0) # /page/1 -> /
        self._add_route(base_redirect)

        page_redirect = RedirectEndpoint(self, prefix_components + [PAGE_URL], endpoint_home_0) # /page -> /
        self._add_route(page_redirect)

        for endpoint in pag.gen_all_pages(endpoint_home_0):
            self._add_route(endpoint)


    def _generate_home_pagination_routes(self):
        # base site
        per_page = self.config.get('per_page', self.DEFAULT_PER_PAGE)
        self._generate_home_pagination_routes_helper(per_page, self.page_store, [])

        # subsites
        for subsite_data in self.subsite_data:
            config = subsite_data['config']
            per_page = config.get('per_page', self.DEFAULT_PER_PAGE)

            # TODO is this a good way to handle this? Should I always combine these for both base site and subsites,
            # and maybe just exclude pages from home listing explicitly by using a pseudo taxonomy or something rather
            # relying on presence/absence of PK?
            page_store = subsite_data.get('page_store', PageStore())

            # message archive doesn't have any PK pages, so just combine these for now. Not sure if this is good default behavior for every subsite though
            page_store += subsite_data.get('non_pk_page_store', PageStore())

            prefix_components = self._resolve_url_components(config['subsite_slug'])

            self._generate_home_pagination_routes_helper(per_page, page_store, prefix_components, config)


    def _generate_404_route(self):
        endpoint = Http404Endpoint(self)
        self._add_route(endpoint)


    def _generate_rss_route(self):
        endpoint = RssEndpoint(self)
        self._add_route(endpoint)


    def _generate_sitemap_route(self):
        endpoint = SiteMapEndpoint(self)
        self._add_route(endpoint)


    def _build_static_file_remapping(self):
        static_file_mapping = {}

        for f in self.static_path.glob('**/*.*'):
            srp = f.relative_to(self.static_path)

            if f.suffix in self.CACHE_BUST_STATIC_EXTENSIONS:
                use_srp = self.wisdom.get_static_file_name(srp)
            else:
                use_srp = srp

            move_to = self.static_output_path / use_srp
            static_file_mapping[srp] = {'from': f, 'to': move_to, 'srp': use_srp}

            components = [self.STATIC_URL] + list(use_srp.parts)
            endpoint = StaticServeEndpoint(self, components)
            self._add_route(endpoint)

        self.static_file_mapping = static_file_mapping


    def static_url(self, srp): # used by templates
        use_srp = self.static_file_mapping[Path(srp)]['srp']
        return f'/{self.STATIC_URL}/{use_srp}'


    def _copy_static(self):
        for meta in self.static_file_mapping.values():
            meta['to'].parent.mkdir(parents=True, exist_ok=True)
            shutil.copyfile(meta['from'], meta['to'])


    def _build_resized_img_endpoints(self):
        for img_data in self.wisdom.gen_cached_images():
            components = img_data['serve_path'].parts[1:] # drop leading '/'
            endpoint = CachedImgEndpoint(self, components)
            self._add_route(endpoint)

            if img_data['copy_original']:
                components = img_data['original_serve_path'].parts[1:] # drop leading '/'
                endpoint = CachedImgEndpoint(self, components)
                self._add_route(endpoint)


    def _copy_cached_imgs(self):
        for img_data in self.wisdom.gen_cached_images(build=True):
            publish_path = img_data['publish_path']
            publish_path.parent.mkdir(parents=True, exist_ok=True)
            shutil.copyfile(img_data['cached_path'], publish_path)

            if img_data['copy_original']:
                publish_path = img_data['original_publish_path']
                publish_path.parent.mkdir(parents=True, exist_ok=True)
                shutil.copyfile(img_data['original_path'], publish_path)


    def get_image_url(self, srp_key, max_width, max_height): # helper/convienence relay method
        return self.wisdom.get_image_url_by_key(srp_key, max_width, max_height)[1] # just return serve path


    def get_url_for_pk(self, pk): # used in TagDefs to dynimacially resolve URLS to pages
        slug = self.pk_link_map[int(pk)]
        url_components = self._resolve_url_components(slug)
        url = '/'.join(url_components)
        return f'/{url}/' # add leading/trailing slashes to make it an absolute link


    def generate(self):
        # remove old/stale published content
        shutil.rmtree(self.output_path, ignore_errors=True) # if dir doesn't exist, ignore error

        for endpoint in self.routes.values():
            #print('rendering', endpoint.url_components, endpoint.url, endpoint.identifier, endpoint.get_output_path())
            rendered = endpoint.render()
            if rendered is None:
                continue

            output_path = self.output_path
            endpoint_output_path = endpoint.get_output_path()

            suffix = endpoint_output_path.suffix
            if suffix:
                fname = endpoint_output_path.name
                if suffix == '.xml':
                    rendered = rendered.lstrip() # strip leading whitespace from xml files to avoid XML parsing error
            else:
                output_path = output_path / endpoint_output_path
                fname = 'index.html'

            output_path.mkdir(parents=True, exist_ok=True)

            with open(output_path / fname, 'w') as f:
                f.write(rendered)

        self._copy_static()
        self._copy_cached_imgs()



import re
from dentmark.defs_manager import DefSet
from dentmark import TagDef, OptionalUnique

class TaxonomyDefSet(DefSet):

    PATTERN = re.compile(r'^root\.meta\.taxonomy\..*[^\.]+$')

    @classmethod
    def copy_from_def_set(cls, other_def_set):
        new_obj = cls(other_def_set.tag_set_name)

        new_obj.pre_tag_addresses = other_def_set.pre_tag_addresses
        new_obj.root_def = other_def_set.root_def
        new_obj.tag_dict = other_def_set.tag_dict
        new_obj._is_checked = other_def_set._is_checked
        new_obj.children_relation_dict = other_def_set.children_relation_dict

        return new_obj


    def get_def(self, tag_address):
        m = self.PATTERN.match(tag_address)
        if m:
            return TaxonomyItemTagDef
        else:
            return self.tag_dict.get(tag_address)


    def get_children_relations(self, tag_address):
        if tag_address.startswith('root.meta.taxonomy'):
            return TaxonomyChildrenRelations()
        else:
            return self.children_relation_dict.get(tag_address, {})


class TaxonomyChildrenRelations:
    # A 'dummy' dictionary that always returns the same thing no matter what the key
    def __getitem__(self, item):
        return OptionalUnique(item) # arg doesn't really matter, just use item I guess in case I need to debug off it. Could be None or '' too

    def items(self):
        return [] # causes .items() to do nothing. None of the taxonomy names are required, so don't need to check for missing names


class TaxonomyItemTagDef(TagDef):
    is_context = True

    min_num_text_nodes = 0
    max_num_text_nodes = 1

    parents = [OptionalUnique('root.meta.taxonomy')]

    def __init__(self, tag_name, address, line_no, indent_level, parent, root, order, nth_of_type, trim_left, trim_right, extra_context):
        self.tag_name = tag_name
        super().__init__(tag_name, address, line_no, indent_level, parent, root, order, nth_of_type, trim_left, trim_right, extra_context)


    def process_data(self, data):
        order = None

        has_order_element = False
        has_child_elements = False

        for child in self.children:
            if not child.is_element:
                has_order_element = True
                order = int(child.text)
                if order < 1:
                    raise ValueError
            else:
                has_child_elements = True

        if has_order_element and has_child_elements:
            raise Exception("Cannot have both an order number set AND child taxonomies listed")

        if has_child_elements:
            return self.context
        else:
            return order


    def validate(self):
        try:
            self.get_data()
        except ValueError:
            return f"Tag '{self.tag_name}' expects a positive integer >= 1"
        except Exception as e:
            return f"Tag '{self.tag_name}' {e}"


class MetaTaxonomyContext(TagDef):
    tag_name = 'taxonomy'
    is_context = True

    parents = [OptionalUnique('root.meta')]

    min_num_text_nodes = 0 # don't allow text nodes
    max_num_text_nodes = 0

    def process_data(self, data):

        taxonomies_dict = {}

        def walk_dict(d, slug_path = ''):
            for k,v in d.items():
                new_slug_path = slug_path
                if new_slug_path:
                    new_slug_path += '/'
                new_slug_path += k

                if type(v) is dict:
                    walk_dict(v, new_slug_path)
                else:
                    taxonomies_dict[new_slug_path] = v

        walk_dict(self.context)

        return taxonomies_dict




import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

setuptools.setup(
    name="indentgen",
    version="0.0.3",
    author="Ryli Dunlap",
    author_email="ryli@transvec.com",
    description="A static site generator for Dentmark",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://www.transvec.com/indentgen",
    packages=setuptools.find_packages(),
    scripts = ['bin/indentgen'],
    install_requires=[
        'dentmark',
        'mako',
        'Pillow'
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
    ],
    python_requires='>=3.6',
)



import datetime

from dentmark import defs_manager, TagDef, Optional, OptionalUnique, RequiredUnique

from dentmark.default_definitions.anchor import Anchor, URLContext, AnchorTitleContext
from dentmark.default_definitions.images import ImgAltContext, ImgTitleContext
from dentmark.default_definitions import Paragraph, Root, BlockQuote, Break
from dentmark.default_definitions.emphasis import Italic
from dentmark.default_definitions.annotation import Annotation, FootNote
from dentmark.default_definitions.lists import UnorderedList, ListItem
from indentgen.default_definitions.config_tag_defs import ConfigURLContext
from indentgen.taxonomy_def_set import MetaTaxonomyContext

CONTENT_TAG_SET = 'indentgen_content'

content_tag_set = defs_manager.copy_tag_set(CONTENT_TAG_SET) # copies from default Dentmark tags by default

content_tag_set.register_tag(MetaTaxonomyContext)


@content_tag_set.register(replace=True)
class IndentgenContentRoot(TagDef):
    tag_name = 'root'

    def before_render(self):
        inline_sum = self.collectors.get('sum')
        meta_summary = self.context['meta'].get('summary')
        pk = self.context['meta'].get('pk')

        taxonomy = self.context['meta'].get('taxonomy', {})

        if not pk and (inline_sum or meta_summary):
            return 'Content with no pk cannot have inline sum tags or meta summary'

        if (inline_sum and meta_summary):
            return 'Cannot declare meta.summary AND inline sum tags. Remove the meta.summary or the inline sum tags.'

        if pk and not (inline_sum or meta_summary):
            return 'Content with pk must declare an inline summary with sum tags OR a meta summary'


        # make sure bookmarks have bm tag
        if 'types/bookmarks' in taxonomy:
            if 'bm' not in self.context['meta']:
                return "Content belonging to taxonomy 'types/bookmarks' must define the 'bm' tag (root.meta.bm)"

            bm = self.context['meta']['bm']
            pub = bm['pub']
            author = bm['author']
            title = bm['title']

            # set the lead if it isn't declared
            if not self.context['meta'].get('lead'):
                self.context['meta']['lead'] = [f'By {author} / {pub}']
                self.context['meta']['lead_content'] = f'By {author} / {pub}'

            # set the description if it isn't declared
            if not self.context['meta'].get('description'):
                self.context['meta']['description'] = f'Thoughts on "{title}" from "{pub}"'

            # set the title if it isn't declared
            if not self.context['meta'].get('title'):
                self.context['meta']['title'] = title

            # TODO: maybe a big change/refactor in the future would be to add the ability to use entirely different TagSets
            # for different types (basically Hugo's notion of a archetype or whatever)

        # promote orphaned text nodes to 'p'
        for i,child in enumerate(self.children):
            if not child.is_element:
                promoted_elem = child.promote(Paragraph)
                self.children[i] = promoted_elem

        # decorate last 'p' with endmark
        # use gen_tags_by_name to find nested p tags (last p tag might be in summary for short posts)
        if self.context['meta'].get('pk') is not None: # Don't do this for info pages that don't have a pk
            p_tags = list(self.gen_tags_by_name('p'))
            if p_tags:
                last_child = self.children[-1]

                # this check ensures that the endmark isn't added if the last child is an img or some other content that
                # would cause the endmark to appear prior to it. Just omit the endmark for posts like this
                if last_child.is_element and (last_child.tag_name == 'p' or last_child.tag_name == 'sum'):
                    p_tags[-1].add_class('p__endmark')

        if 'types/articles' in taxonomy or 'types/thoughts' in taxonomy:
            # ensure thoughts and articles have title and description in meta
            # bookmarks are exempts since these are derived from root.meta.bm, but
            # override the derived versions if present

            if 'title' not in self.context['meta']:
                return "Content belonging to taxonomy 'articles' or 'thoughts' must define the 'title' tag (root.meta.title)"

            if 'description' not in self.context['meta']:
                return "Content belonging to taxonomy 'articles' or 'thoughts' must define the 'description' tag (root.meta.description)"


        if 'types/articles' in taxonomy:
            # use dropcap for first p of article and p following h2
            # use gen_tags_by_name to find nested p tags (first p tag might be in summary)

            # dropcap to first p in article, even if it is nested in a sum tag
            for child in self.gen_tags_by_name('p'):
                child.context['use_dropcap'] = ''
                break

            # add dropcaps to the first p following an h2
            for child in self.gen_tags_by_name('h2'):
                next_p = child.next_tag_of_type('p')
                if next_p:
                    next_p.context['use_dropcap'] = ''

            # add leadin styling to the first p following an h3
            for child in self.gen_tags_by_name('h3'):
                next_p = child.next_tag_of_type('p')
                if next_p:
                    next_p.context['use_leadin'] = ''

        # make sure only photos types have gallery tag
        if 'gallery' in self.context['meta']:
            # check against all the taxonomies since taxonomies specify gallery
            gallery_valid = False
            tax_defs = self.extra_context['indentgen'].taxonomy_map
            for tax_slug_path in taxonomy:
                if tax_defs[tax_slug_path]['gallery']:
                    gallery_valid = True
                    break
            if not gallery_valid:
                return "Only content with a taxonomy with 'gallery' set may define the 'gallery' tag (root.meta.gallery)"

            # do not specify meta.img since galleries draw from meta.gallery.img OR use the 1st image in the gallery
            # have to do it this way since the img tag won't populate until after render
            meta_def = self.get_child_by_name('meta')
            if meta_def.get_child_by_name('img'):
                return "meta.img prohibited for a gallery page. These draw the featured image from meta.gallery.img OR use the 1st image in the gallery"

            # check that multiple covers aren't specified
            if len(self.collectors.get('cover', [])) > 1:
                return "Only 1 cover tag can be specified for a root.meta.gallery.img tag. Multiple cover tags detected"


    def render_main(self):
        body = f'{self.content}'

        fns = self.collectors.get('fn', [])
        fns_rendered = ''.join(fns)

        taxonomy = self.context['meta'].get('taxonomy', {})

        if 'bm' in self.collectors:
            body = self.collectors['bm'][0] + body

        if 'articles' in taxonomy or self.context['meta']['slug'] == 'about':
            indentgen = self.extra_context['indentgen']

            if 'creative-writing' in taxonomy:
                url = '/who-is-nally-dupri/'
                sig_suffix = 'nd'
                alt = 'Nally Duprí'
            else:
                url = '/about/'
                sig_suffix = 'rd'
                alt = 'Ryli Dunlap'

            body += f'<p><a href="{url}"><img class="signature" src="/{indentgen.STATIC_URL}/img/signature_{sig_suffix}.png" alt="{alt}" /></a></p>'

        if fns_rendered:
            body += f'<section class="footnotes" role="doc-endnotes"><hr/><ol>{fns_rendered}</ol></section>'

        return body


@content_tag_set.register(replace = True)
class IndentgenContentParagraph(Paragraph):

    parents = [Optional('root'), Optional('root.sum'), Optional('root.meta.summary'), Optional('root.bq'), Optional('root.conv.msg'), Optional('root.p.a8n.fn.bq')]

    def render_main(self):
        classes = ''
        if self.classes:
            class_str = ' '.join(self.classes)
            classes = f' class="{class_str}"'

        dropcap = ''
        first_letter = self.context.get('use_dropcap')

        if first_letter:
            dropcap = f'<span class="dropcap">{first_letter}</span>'

        leadin = ''
        leadin_txt = self.context.get('use_leadin')

        if leadin_txt:
            commaspace = '' if self.content[0] == ',' else ' '
            leadin = f'<span class="leadin">{leadin_txt}</span>{commaspace}'

        return f'<p{classes}>{dropcap}{leadin}{self.content}</p>'


    def before_render(self):
        # This has to be done here prior to render, so it take effect

        text_nodes = self.strip_tags()

        if 'use_dropcap' in self.context:
            first_letter = None
            if text_nodes:
                first_letter = text_nodes[0].text[0] if text_nodes[0].text else None
                if first_letter:
                    text_nodes[0].text = text_nodes[0].text[1:] # strip first letter
                    self.context['use_dropcap'] = first_letter

        if 'use_leadin' in self.context:
            if text_nodes:
                first_part = text_nodes[0].text if text_nodes[0].text else None
                if first_part:
                    # only go up to comma if it is present
                    before_comma = first_part.split(',')[0]
                    words = before_comma.split(' ')[:5] # max number of words for lead-in
                    rejoin = ' '.join(words)
                    # trim the leadin text off the text node
                    text_nodes[0].text = first_part[len(rejoin):]
                    self.context['use_leadin'] = rejoin


@content_tag_set.register()
class MetaContext(TagDef):
    tag_name = 'meta'
    is_context = True

    min_num_text_nodes = 0
    max_num_text_nodes = 0

    parents = [RequiredUnique('root')]

    def process_data(self, data):
        return self.context


@content_tag_set.register()
class TitleMetaContext(TagDef):
    tag_name = 'title'
    is_context = True

    min_num_text_nodes = 1
    max_num_text_nodes = 1

    parents = [OptionalUnique('root.meta')]


@content_tag_set.register()
class IndentgenContentMetaUseTemplate(TagDef):
    tag_name = 'use_template'
    is_context = True

    min_num_text_nodes = 1
    max_num_text_nodes = 1

    parents = [OptionalUnique('root.meta')]

    def validate(self):
        path = self.get_data()
        indentgen = self.extra_context['indentgen']
        try:
            indentgen.templates.get_template(path)
        except Exception as e:
            return f'use_template: {e}'


# register this separately since TitleMetaContext is imported and used for
# Taxonomy config definitons too, which doesn't use root.meta.bm
@content_tag_set.register()
class TitleMetaBMContext(TitleMetaContext):
    parents = [RequiredUnique('root.meta.bm')]


@content_tag_set.register()
class SlugContext(ConfigURLContext):
    tag_name = 'slug'

    parents = [RequiredUnique('root.meta')]


@content_tag_set.register()
class PKContext(TagDef):
    tag_name = 'pk'
    is_context = True

    min_num_text_nodes = 1
    max_num_text_nodes = 1

    parents = [OptionalUnique('root.meta')]

    def process_data(self, data):
        if not data:
            return None
        pk = int(data[0])
        if pk < 1:
            raise ValueError
        return pk


    def validate(self):
        try:
            self.get_data()
        except ValueError:
            return f"Tag '{self.tag_name}' expects a positive integer >= 1"


@content_tag_set.register()
class MetaDateContext(TagDef):
    tag_name = 'date'
    is_context = True

    min_num_text_nodes = 1
    max_num_text_nodes = 1

    parents = [OptionalUnique('root.meta')]

    date_format = '%Y-%m-%d-%H%M'

    def process_data(self, data):
        time = data[0].strip()
        if len(time) == 10:
            # add default time
            time += '-0000'

        return datetime.datetime.strptime(time, self.date_format)

    def validate(self):
        try:
            # calls process_data() on first child text node
            self.get_data()
        except ValueError:
            return f"Invalid date value for Tag '{self.tag_name}' in meta"


@content_tag_set.register()
class MetaDisableCommentsContext(TagDef):
    tag_name = 'disable_comments'
    is_context = True

    min_num_text_nodes = 0
    max_num_text_nodes = 1

    parents = [OptionalUnique('root.meta')]

    def process_data(self, data):
        if data:
            return data[0].lower() == 'true'
        return True

    def validate(self):
        if self.children:
            if self.children[0].get_data().lower() not in ('true', 'false'):
                return f"'{self.tag_name}' tag value must be either 'true', 'false', or [empty]. Defaults to 'true' if [empty]"


@content_tag_set.register()
class ContentLead(TagDef):
    tag_name = 'lead'
    is_context = True

    parents = [OptionalUnique('root.meta')]

    def render_main(self):
        self.parent.context[f'{self.tag_name}_content'] = self.content
        return ''


@content_tag_set.register()
class ContentMetaDescription(TagDef):
    tag_name = 'description'
    is_context = True

    min_num_text_nodes = 1
    max_num_text_nodes = 1

    parents = [OptionalUnique('root.meta')]


# Patch Image tag
@content_tag_set.register(replace=True)
class IndentgenContentImage(TagDef):
    tag_name = 'img'

    min_num_text_nodes = 1
    max_num_text_nodes = 1

    parents = [Optional('root')]

    MAX_WIDTH = 800
    MAX_HEIGHT = 600

    def get_image_data(self):
        srp = self.extra_context['srp']
        wisdom = self.extra_context['indentgen'].wisdom
        lto = self.context.get('link_to_original', False)
        return wisdom.get_image_url(srp, self.content, self.MAX_WIDTH, self.MAX_HEIGHT, copy_original=lto)

    def render_main(self):
        key, resized_serve_path, original_serve_path = self.get_image_data()

        alt = self.context.get('alt', '')
        caption = self.context.get('caption', [])

        use_alt_text = alt or ' '.join(caption)
        use_alt = f' alt="{use_alt_text}"' if use_alt_text else ''

        img_src = f'<img src="{resized_serve_path}"{use_alt} />'

        lto = self.context.get('link_to_original', False)

        if lto:
            img_src = f'<a href="{original_serve_path}">{img_src}</a>'

        caption_content = self.context.get('caption_content', '')
        attr_content = self.context.get('attr_content', '')

        caption_attr_src = caption_content
        if caption_content and attr_content:
            caption_attr_src += ' / '
        caption_attr_src += attr_content

        if caption_attr_src:
            caption_attr_src = f'<figcaption><p class="figcaption__content">{caption_attr_src}</p></figcaption>'

        return f'<figure>{img_src}{caption_attr_src}</figure>'


@content_tag_set.register()
class IndentgenContentImageMeta(IndentgenContentImage):

    parents = [OptionalUnique('root.meta'), Optional('root.meta.gallery')]

    def render_main(self):
        img_ctx = self.parent.context.setdefault('img', [])
        self.context['key'] = self.get_image_data()[0]
        img_ctx.append(self.context)
        return ''


@content_tag_set.register()
class IndentgenContentImageLinkToOriginal(MetaDisableCommentsContext):
    tag_name = 'link_to_original'

    parents = [OptionalUnique('root.img')]


@content_tag_set.register()
class IndentgenContentImageAttr(TagDef):
    tag_name = 'attr'
    is_context = True

    parents = [OptionalUnique('root.img'), OptionalUnique('root.meta.img'), OptionalUnique('root.meta.gallery.img')]

    def render_main(self):
        self.parent.context[f'{self.tag_name}_content'] = self.content
        return ''


@content_tag_set.register()
class IndentgenContentImageCaption(IndentgenContentImageAttr):
    tag_name = 'caption'
    is_context = True

    def process_data(self, data):
        return [' '.join(x.text for x in self.strip_tags())]


@content_tag_set.register()
class IndentgenContentImageGalleryCover(MetaDisableCommentsContext):
    tag_name = 'cover'
    add_to_collector = True

    parents = [OptionalUnique('root.meta.gallery.img')]


@content_tag_set.register()
class IndentgenContentSummaryBody(TagDef):
    tag_name = 'sum'
    add_to_collector = True

    parents = [Optional('root'), Optional('root.p'), Optional('root.p.a8n')]


@content_tag_set.register()
class IndentgenContentSummaryMeta(TagDef):
    tag_name = 'summary'
    is_context = True

    parents = [OptionalUnique('root.meta')]

    def validate(self):
        if not self.children:
            return 'Meta summary cannot be blank/empty if present. Must have at least one child'

    def render_main(self):
        self.parent.context[f'{self.tag_name}_content'] = self.content
        return ''


# patch these for img
@content_tag_set.register()
class IndentgenContentImageMetaAlt(ImgAltContext):
    parents = [Optional('root.meta.img')]


@content_tag_set.register()
class IndentgenContentImageMetaTitle(ImgTitleContext):
    parents = [Optional('root.meta.img')]


@content_tag_set.register(replace=True)
class IndentgenContentImageAnchor(Anchor):
    parents = [
        Optional('root'),
        Optional('root.p'),
        Optional('root.p.a8n.fn'),
        Optional('root.bq.a8n.fn'),
        Optional('root.img.attr'),
        Optional('root.meta.img.attr'),
        Optional('root.p.sum'),
        Optional('root.p.sum.i'),
        Optional('root.sum.p'),
        Optional('root.ul.li'),
        Optional('root.p.a8n.fn.i'),
        Optional('root.p.i'),
        Optional('root.img.caption'),
        Optional('root.conv.msg.p'),
        Optional('root.bq.p.a8n.fn'),
        Optional('root.ul.li.a8n.fn.i'),
        Optional('root.aside'),
        Optional('root.extract'),
        Optional('root.p.hl')
    ]


    def render_main(self):
        url = self.context.get('url')
        if url is None:
            if self.content.startswith('http'):
                url = self.content
            else:
                content_split = self.content.split('@')
                if len(content_split) == 2 and '.' in content_split[1]:
                    url = f'mailto:{self.content}'
                else:
                    # don't add the tag. This supports the #TODO or 'deferred' links
                    return self.content

        href = f' href="{url}"' if url else ''

        title = self.context.get('title')
        title_str = f' title="{title}"' if title else ''

        return f'<a{href}{title_str}>{self.content}</a>'


@content_tag_set.register(replace=True)
class IndentgenContentImageAnchorURL(URLContext):
    parents = [
        OptionalUnique('root.a'),
        OptionalUnique('root.p.a'),
        Optional('root.p.a8n.fn.a'),
        Optional('root.bq.a8n.fn.a'),
        OptionalUnique('root.img.attr.a'),
        OptionalUnique('root.meta.img.attr.a'),
        OptionalUnique('root.p.sum.i.a'),
        OptionalUnique('root.p.a8n.fn.i.a'),
        OptionalUnique('root.p.i.a'),
        OptionalUnique('root.p.sum.a'),
        OptionalUnique('root.conv.msg.p.a'),
        OptionalUnique('root.bq.p.a8n.fn.a'),
        OptionalUnique('root.ul.li.a8n.fn.i.a'),
        OptionalUnique('root.aside.a'),
        OptionalUnique('root.ma'),
        OptionalUnique('root.extract.a'),
        OptionalUnique('root.p.hl.a')
    ]


    def process_data(self, data):
        if data[0].isdigit():
            indentgen = self.extra_context['indentgen']
            return indentgen.get_url_for_pk(data[0])
        return data[0]

    def validate(self):
        raw_val = self.children[0].text
        try:
            self.get_data()
        except KeyError as e:
            return f'Page with PK of {raw_val} does not exist'


@content_tag_set.register()
class IndentgenContentImageAnchorTitle(AnchorTitleContext):
    parents = [OptionalUnique('root.img.attr.a'), OptionalUnique('root.meta.img.attr.a')]


@content_tag_set.register()
class IndentgenContentSumItalic(Italic):
    parents = [
        Optional('root.p.sum'),
        Optional('root.bq.p'),
        Optional('root.p.a8n.fn'),
        Optional('root.p.a'),
        Optional('root.p.a8n.fn.a'),
        Optional('root.bq.p.a8n.fn'),
        Optional('root.p.pq'),
        Optional('root.p.a8n'),
        Optional('root.img.caption'),
        Optional('root.ul.li.a8n.fn')
    ]


@content_tag_set.register()
class IndentgenContenBqA8n(Annotation):
    parents = [Optional('root.bq.p'), Optional('root.ul.li')]

@content_tag_set.register()
class IndentgenContenBqFootnote(FootNote):
    parents = [Optional('root.bq.p.a8n'), Optional('root.ul.li.a8n')]

@content_tag_set.register()
class IndentgenContentBr(Break):
    parents = [Optional('root.conv.msg.p')]


@content_tag_set.register(replace=True)
class IndentgenContentBlockQuote(BlockQuote):

    def before_render(self):
        # promote orphaned text nodes to 'p'
        for i,child in enumerate(self.children):
            if not child.is_element:
                promoted_elem = child.promote(Paragraph)
                self.children[i] = promoted_elem


@content_tag_set.register(replace=True)
class IndentgenContentUnorderedList(UnorderedList):

    # the default dentmark definitions don't allow text-only nodes,
    # but since we're promoting them in indentgen, re-set these
    # defaults.
    min_num_text_nodes = 0
    max_num_text_nodes = None

    def before_render(self):
        # promote orphaned text nodes to 'li'
        for i,child in enumerate(self.children):
            if not child.is_element:
                promoted_elem = child.promote(ListItem)
                self.children[i] = promoted_elem


@content_tag_set.register()
class IndentgenContentPullQuote(TagDef):
    tag_name = 'pq'
    add_to_collector = True

    min_num_text_nodes = 1
    max_num_text_nodes = None

    parents = [Optional('root.p')]

    def render_secondary(self):
        pass

    def render_secondary(self):
        content = self.content

        if content:
            # make sure first letter is capitalized, so that we can pull fragments of sentences as pullquotes
            content = content[0].upper() + content[1:]

        return f'<div class="pullquote"><span class="openquote"></span><span class="quotecontent">{content}</span></div>'


@content_tag_set.register()
class IndentgenContentPutRef(TagDef):
    tag_name = 'putref'

    min_num_text_nodes = 1
    max_num_text_nodes = 1

    parents = [Optional('root')]

    # Use get_data() here to get the value since self.content won't be available yet during the before_render check

    def render_main(self):
        id_split = self.get_data().split('-')
        index = int(id_split[-1]) - 1 # tag_value is 1-indexed
        if index < 0:
            raise ValueError
        tag_name_key = '-'.join(id_split[:-1])
        return self.root.collectors[tag_name_key][index]

    def before_render(self):
        try:
            self.render_main()
        except (ValueError, KeyError, IndexError) as e:
            return f'Invalid putref key: {self.get_data()}'


@content_tag_set.register()
class IndentgenContentBookMark(TagDef):
    tag_name = 'bm'
    is_context = True
    add_to_collector = True

    min_num_text_nodes = 0
    max_num_text_nodes = 0

    parents = [OptionalUnique('root.meta')]

    def process_data(self, data):
        return self.context

    def render_secondary(self):
        pub = self.context['pub']
        url = self.context['url']
        author = self.context['author']
        diigo = self.context.get('diigo')

        annotated_src = ''
        if diigo:
            annotated_src = f' <small>[ <a href="https://diigo.com/{diigo}">Annotated Version</a> ]</small>'

        return f'<p class="bookmark__link">Original Source: <a href="{url}"><i>{pub}</i></a>{annotated_src}</p>'


@content_tag_set.register()
class IndentgenContentBookMarkURL(TagDef):
    tag_name = 'url'
    is_context = True

    min_num_text_nodes = 1
    max_num_text_nodes = 1

    parents = [RequiredUnique('root.meta.bm')]


@content_tag_set.register()
class IndentgenContentBookMarkPublication(IndentgenContentBookMarkURL):
    tag_name = 'pub'


@content_tag_set.register()
class IndentgenContentBookMarkAuthor(IndentgenContentBookMarkURL):
    tag_name = 'author'


@content_tag_set.register()
class IndentgenContentBookMarkDiigo(TagDef):
    tag_name = 'diigo'
    is_context = True

    min_num_text_nodes = 1
    max_num_text_nodes = 1

    parents = [OptionalUnique('root.meta.bm')]


@content_tag_set.register()
class IndentgenContentConversation(TagDef):
    tag_name = 'conv'

    parents = [Optional('root')]


@content_tag_set.register()
class IndentgenContentConversationDefault(TagDef):
    tag_name = 'default_from'
    is_context = True

    min_num_text_nodes = 1
    max_num_text_nodes = 1

    parents = [OptionalUnique('root.conv')]


@content_tag_set.register()
class IndentgenContentMsg(TagDef):
    tag_name = 'msg'

    parents = [Optional('root.conv')]

    def before_render(self):
        # promote orphaned text nodes to 'li'
        for i,child in enumerate(self.children):
            if not child.is_element:
                promoted_elem = child.promote(Paragraph)
                self.children[i] = promoted_elem


    def render_main(self):
        msg_str = ''

        actor_id_map = self.parent.context.setdefault('msg_actors', {})

        prev_actor_id = self.parent.context.get('prev_actor_id')

        default_actor_name = self.parent.context.get('default_from') or 'Me' # default name for messages I sent (ones without a 'from' context)

        #for msg in msgs:
        frm = self.context.get('from')
        content = self.content

        side = 'left' if frm else 'right'
        cls_color = ''

        put_name = ''

        actor_id = -1 # denotes 'Me'

        if frm:
            actor_id = actor_id_map.setdefault(frm, len(actor_id_map))
            cls_color = f' msg__color__{actor_id}'

        if actor_id != prev_actor_id:
            use_name = frm or default_actor_name
            put_name = f'<span>{use_name}</span>'
            self.parent.context['prev_actor_id'] = actor_id

        return f'<div class="conversation__msg msg__{side}{cls_color}">{put_name}{content}</div>'


@content_tag_set.register()
class IndentgenContentMsgFrom(TagDef):
    tag_name = 'from'
    is_context = True

    min_num_text_nodes = 1
    max_num_text_nodes = 1

    parents = [OptionalUnique('root.conv.msg')]


@content_tag_set.register()
class IndentgenContentAside(TagDef):
    tag_name = 'aside'

    parents = [Optional('root')]

    def render_main(self):
        return f'<div class="aside">{self.content}</div>'


@content_tag_set.register()
class IndentgenContentMessageArchiveLink(TagDef):
    tag_name = 'ma'

    min_num_text_nodes = 1
    max_num_text_nodes = 1

    parents = [Optional('root')]

    def render_main(self):
        url = self.context.get('url')

        if url:
            return f'<a href="{url}" class="message-archive-link">{self.content}</a>'

        return ''


@content_tag_set.register()
class IndentgenContentMessageExtract(TagDef):
    tag_name = 'extract'

    parents = [Optional('root')]

    def render_main(self):
        link = self.context.get('link')
        return f'<blockquote>{self.content}<sup><a href="{link}" class="extract-ref">↩︎</a></sup></blockquote>'


@content_tag_set.register()
class IndentgenContentMessageExtractLink(IndentgenContentImageAnchorURL):
    tag_name = 'link'

    parents = [RequiredUnique('root.extract')]


@content_tag_set.register()
class IndentgenContentPhotoGallery(TagDef):
    tag_name = 'gallery'
    is_context = True

    min_num_text_nodes = 0
    max_num_text_nodes = 0

    parents = [OptionalUnique('root.meta')]

    def process_data(self, data):
        return self.context


@content_tag_set.register()
class IndentgenContentPhotoGalleryPerPage(TagDef):
    tag_name = 'per_page'
    is_context = True

    min_num_text_nodes = 1
    max_num_text_nodes = 1

    parents = [OptionalUnique('root.meta.gallery')]

    def process_data(self, data):
        per_page = int(data[0])
        if per_page < 0:
            raise ValueError
        return per_page


    def validate(self):
        try:
            self.get_data()
        except ValueError:
            return f"Tag '{self.tag_name}' in meta.gallery expects a positive integer"

